@@ -15,13 +15,13 @@
 import base64
 import json
 import os
+import pathlib
 import re
-import shlex
 import shutil
-import subprocess
 
 import click
 from tfworker.commands.base import BaseCommand
+from tfworker.util.system import pipe_exec
 
 
 class HookError(Exception):
@@ -46,10 +46,10 @@ class TerraformCommand(BaseCommand):
 
         self._b64_encode = kwargs.get("b64_encode")
         self._deployment = kwargs.get("deployment")
-        self._force_apply = kwargs.get("force_apply")
+        self._force = kwargs.get("force")
         self._show_output = kwargs.get("show_output")
         self._terraform_bin = kwargs.get("terraform_bin")
-
+        self._terraform_modules_dir = kwargs.get("terraform_modules_dir")
         self._plan_for = "destroy" if kwargs.get("destroy") else "apply"
         (self._tf_version_major, self._tf_version_minor) = kwargs.get(
             "tf_version", (None, None)
@@ -59,15 +59,42 @@ class TerraformCommand(BaseCommand):
                 self._tf_version_major,
                 self._tf_version_minor,
             ) = self.get_terraform_version(self._terraform_bin)
-        super(TerraformCommand, self).__init__(rootc, plan_for=self._plan_for, **kwargs)
+        super(TerraformCommand, self).__init__(
+            rootc,
+            plan_for=self._plan_for,
+            tf_version_major=self._tf_version_major,
+            **kwargs,
+        )
 
     @property
     def plan_for(self):
         return self._plan_for
 
+    @property
+    def tf_version_major(self):
+        return self._tf_version_major
+
     def prep_modules(self):
         """Puts the modules sub directories into place."""
-        mod_source = f"{self._repository_path}/terraform-modules".replace("//", "/")
+
+        if self._terraform_modules_dir:
+            mod_source = self._terraform_modules_dir
+            mod_path = pathlib.Path(mod_source)
+            if not mod_path.exists():
+                click.secho(
+                    f'The specified terraform-modules directory "{mod_source}" does not exists',
+                    fg="red",
+                )
+                raise SystemExit(1)
+        else:
+            mod_source = f"{self._repository_path}/terraform-modules".replace("//", "/")
+            mod_path = pathlib.Path(mod_source)
+            if not mod_path.exists():
+                click.secho(
+                    "The terraform-modules directory does not exist.  Skipping.",
+                    fg="green",
+                )
+                return
         mod_destination = f"{self._temp_dir}/terraform-modules".replace("//", "/")
         click.secho(
             f"copying modules from {mod_source} to {mod_destination}", fg="yellow"
@@ -80,7 +107,13 @@ class TerraformCommand(BaseCommand):
         )
 
     def exec(self):
-        for definition in self.definitions.limited():
+        try:
+            def_iter = self.definitions.limited()
+        except ValueError as e:
+            click.secho(f"Error with supplied limit: {e}", fg="red")
+            raise SystemExit(1)
+
+        for definition in def_iter:
             execute = False
             # copy definition files / templates etc.
             click.secho(f"preparing definition: {definition.tag}", fg="green")
@@ -94,7 +127,10 @@ class TerraformCommand(BaseCommand):
                 click.secho("error running terraform init", fg="red")
                 raise SystemExit(1)
 
-            click.secho(f"planning definition: {definition.tag}", fg="green")
+            click.secho(
+                f"planning definition for {self._plan_for}: {definition.tag}",
+                fg="green",
+            )
 
             # run terraform plan
             try:
@@ -105,6 +141,11 @@ class TerraformCommand(BaseCommand):
                     plan_action=self._plan_for,
                 )
             except PlanChange:
+                # on destroy, terraform ALWAYS indicates a plan change, at least as of
+                # version 0.12.x, unsure why
+                click.secho(
+                    f"plan changes for {self._plan_for} {definition.tag}", fg="red"
+                )
                 execute = True
             except TerraformError:
                 click.secho(
@@ -113,27 +154,19 @@ class TerraformCommand(BaseCommand):
                 )
                 raise SystemExit(2)
 
-            if self._force_apply:
-                execute = True
-
-            if execute and self._tf_apply:
-                if self._force_apply:
-                    click.secho(
-                        f"force apply for {definition.tag}, applying",
-                        fg="yellow",
-                    )
-                else:
-                    click.secho(
-                        f"plan changes for {definition.tag}, applying",
-                        fg="yellow",
-                    )
-            elif execute and self._destroy:
-                click.secho(
-                    f"plan changes for {definition.tag}, destroying",
-                    fg="yellow",
-                )
-            elif not execute:
+            if not execute:
                 click.secho(f"no plan changes for {definition.tag}", fg="yellow")
+
+            if self._force and (self._tf_apply or self._destroy):
+                execute = True
+                click.secho(f"forcing {self._plan_for} due to --force", fg="red")
+
+            # there are no plan changes, and no forced execution, so move on
+            if not execute:
+                continue
+
+            # no apply/destroy requested, so move on
+            if not self._tf_apply and not self._destroy:
                 continue
 
             try:
@@ -157,7 +190,7 @@ class TerraformCommand(BaseCommand):
 
     def _run(self, definition, command, debug=False, plan_action="init"):
         """Run terraform."""
-        if self._tf_version_major == 12:
+        if self._tf_version_major >= 12:
             params = {
                 "init": f"-input=false -no-color -plugin-dir={self._temp_dir}/terraform-plugins",
                 "plan": "-input=false -detailed-exitcode -no-color",
@@ -220,7 +253,7 @@ class TerraformCommand(BaseCommand):
         click.secho(
             f"cmd: {self._terraform_bin} {command} {command_params}", fg="yellow"
         )
-        (exit_code, stdout, stderr) = TerraformCommand.pipe_exec(
+        (exit_code, stdout, stderr) = pipe_exec(
             f"{self._terraform_bin} {command} {command_params}",
             cwd=working_dir,
             env=env,
@@ -362,7 +395,7 @@ class TerraformCommand(BaseCommand):
             )
 
         # execute the hook
-        (exit_code, stdout, stderr) = TerraformCommand.pipe_exec(
+        (exit_code, stdout, stderr) = pipe_exec(
             f"{hook_script} {phase} {command}",
             cwd=hook_dir,
             env=local_env,
@@ -379,75 +412,6 @@ class TerraformCommand(BaseCommand):
         if exit_code != 0:
             raise HookError("hook script {}")
 
-    @staticmethod
-    def pipe_exec(args, stdin=None, cwd=None, env=None):
-        """
-        A function to accept a list of commands and pipe them together.
-
-        Takes optional stdin to give to the first item in the pipe chain.
-        """
-        count = 0
-        commands = []
-        if env is None:
-            env = os.environ.copy()
-
-        if not isinstance(args, list):
-            args = [args]
-
-        for i in args:
-            if count == 0:
-                if stdin is None:
-                    commands.append(
-                        subprocess.Popen(
-                            shlex.split(i),
-                            stdout=subprocess.PIPE,
-                            stderr=subprocess.PIPE,
-                            cwd=cwd,
-                            env=env,
-                        )
-                    )
-                else:
-                    commands.append(
-                        subprocess.Popen(
-                            shlex.split(i),
-                            stdout=subprocess.PIPE,
-                            stderr=subprocess.PIPE,
-                            stdin=subprocess.PIPE,
-                            cwd=cwd,
-                            env=env,
-                        )
-                    )
-            else:
-                commands.append(
-                    subprocess.Popen(
-                        shlex.split(i),
-                        stdout=subprocess.PIPE,
-                        stderr=subprocess.PIPE,
-                        stdin=commands[count - 1].stdout,
-                        cwd=cwd,
-                        env=env,
-                    )
-                )
-            count = count + 1
-
-        if stdin is not None:
-            stdin_bytes = stdin.encode()
-            if len(commands) > 1:
-                commands[0].communicate(input=stdin_bytes)
-                stdout, stderr = commands[-1].communicate()
-                commands[-1].wait()
-                returncode = commands[-1].returncode
-            else:
-                stdout, stderr = commands[0].communicate(input=stdin_bytes)
-                commands[0].wait()
-                returncode = commands[0].returncode
-        else:
-            stdout, stderr = commands[-1].communicate()
-            commands[-1].wait()
-            returncode = commands[-1].returncode
-
-        return (returncode, stdout, stderr)
-
     @staticmethod
     def check_hooks(phase, working_dir, command):
         """
@@ -472,7 +436,7 @@ class TerraformCommand(BaseCommand):
         """
         base_dir, _ = os.path.split(working_dir)
         try:
-            (exit_code, stdout, stderr) = TerraformCommand.pipe_exec(
+            (exit_code, stdout, stderr) = pipe_exec(
                 f"{terraform_bin} output -json -no-color {item}",
                 cwd=f"{base_dir}/{state}",
                 env=env,
@@ -498,9 +462,7 @@ class TerraformCommand(BaseCommand):
 
     @staticmethod
     def get_terraform_version(terraform_bin):
-        (return_code, stdout, stderr) = TerraformCommand.pipe_exec(
-            f"{terraform_bin} version"
-        )
+        (return_code, stdout, stderr) = pipe_exec(f"{terraform_bin} version")
         if return_code != 0:
             click.secho(f"unable to get terraform version\n{stderr}", fg="red")
             raise SystemExit(1)
