@@ -25,4 +25,4 @@ DEFAULT_GCP_REGION = "us-west2b"
 DEFAULT_BACKEND_REGION = "us-west-2"
 DEFAULT_TERRFORM = "/usr/local/bin/terraform"
 
-RESERVED_FILES = ["terraform.tf", "worker-locals.tf"]
+RESERVED_FILES = ["terraform.tf", "worker-locals.tf", "worker.auto.tfvars"]
